import itertools
import bpy
import mathutils

class MeshExporter:
	def __init__(self):
		self.show_progress = True
		self.use_strips = True
		self.use_degen_tris = False

	def join_strips(self, strips):
		big_strip = []

		for s in strips:
			if big_strip:
				# Generate glue elements, ensuring that the next strip begins at
				# an even position
				glue = [big_strip[-1], s[0]]
				if len(big_strip)%2:
					glue += [s[0]]

				big_strip += glue

			big_strip += s

		return big_strip

	def export_to_file(self, context, out_fn):
		obj = context.active_object

		from .util import Progress

		progress = Progress(self.show_progress and context)
		progress.push_task("", 0.0, 0.95)
		resource = self.export_mesh(context, obj, progress)

		with open(out_fn, "w") as out_file:
			for s in resource.statements:
				s.write_to_file(out_file)

	def export_mesh(self, context, mesh_or_obj, progress):
		from .mesh import Mesh, create_mesh_from_object

		if type(mesh_or_obj)==Mesh:
			mesh = mesh_or_obj
		else:
			progress.push_task("", 0.0, 0.9)
			mesh = create_mesh_from_object(context, mesh_or_obj, progress)
			progress.pop_task()

		from .datafile import Resource, Statement, Token
		resource = Resource(mesh.name+".mesh")
		statements = resource.statements

		st = Statement("vertices", Token("NORMAL3"))
		if mesh.uv_layers:
			for u in mesh.uv_layers:
				size = str(len(u.uvs[0]))
				if u.unit==0:
					st.append(Token("TEXCOORD"+size))
				else:
					st.append(Token("TEXCOORD{}_{}".format(size, u.unit)))
			if mesh.tbn_vecs:
				st.append(Token("TANGENT3"))
				st.append(Token("BINORMAL3"))
		if mesh.vertex_groups:
			st.append(Token("ATTRIB{}_5".format(mesh.max_groups_per_vertex*2)))
		st.append(Token("VERTEX3"))

		normal = None
		uvs = [None]*len(mesh.uv_layers)
		tan = None
		bino = None
		group = None
		for v in mesh.vertices:
			if v.normal!=normal:
				st.sub.append(Statement("normal", *v.normal))
				normal = v.normal
			for i, u in enumerate(mesh.uv_layers):
				if v.uvs[i]!=uvs[i]:
					if u.unit==0:
						st.sub.append(Statement("texcoord", *v.uvs[i]))
					else:
						st.sub.append(Statement("multitexcoord", u.unit, *v.uvs[i]))
					uvs[i] = v.uvs[i]
			if mesh.tbn_vecs:
				if v.tan!=tan:
					st.sub.append(Statement("tangent", *v.tan))
					tan = v.tan
				if v.bino!=bino:
					st.sub.append(Statement("binormal", *v.bino))
					bino = v.bino
			if mesh.vertex_groups:
				group_attr = [(group_index_map[g.group], g.weight*v.group_weight_scale) for g in v.groups[:mesh.max_groups_per_vertex]]
				while len(group_attr)<mesh.max_groups_per_vertex:
					group_attr.append((0, 0.0))
				group_attr = list(itertools.chain(*group_attr))
				if group_attr!=group:
					st.sub.append(Statement("attrib", 5, *group_attr))
					group = group_attr
			st.sub.append(Statement("vertex", *v.co))

		statements.append(st)

		if self.use_strips:
			strips = mesh.vertex_sequence
			if self.use_degen_tris:
				strips = [self.join_strips(strips)]

			for s in strips:
				st = Statement("batch", Token("TRIANGLE_STRIP"))
				for i in range(0, len(s), 32):
					st.sub.append(Statement("indices", *(v.index for v in s[i:i+32])))
				statements.append(st)
		else:
			st = Statement("batch", "TRIANGLES")
			for f in mesh.faces:
				st.sub.append(Statement("indices", *(v.index for v in f.vertices)))
			statements.append(st)

		if mesh.lines:
			st = Statement("batch", "LINES")
			for l in mesh.lines:
				st.sub.append(Statement("indices", *(v.index for v in l.vertices)))
			statements.append(st)

		if mesh.winding_test:
			statements.append(Statement("winding", "COUNTERCLOCKWISE"))

		progress.set_progress(1.0)

		return resource
